
\begin{code}
module Databases where

open import Data.Function
open import Data.String
open import Data.Nat
open import Data.Vec
open import Data.List
open import Data.Bool
open import Data.Pair
open import Data.Maybe
open import Foreign.IO
import Control.Monad
open Control.Monad returnIO bindIO

VString : Nat -> Set
VString n = Vec Char n

vString : {n : Nat}(s : String) -> Vec Char (strLen s)
vString = listToVec ∘ stringToList

data FieldType : Set where
  text : Nat -> FieldType
  number : FieldType

Field : FieldType -> Set
Field (text n) = VString n
Field number   = Nat

FieldName = String
TableName = String
TableType = List (FieldName × FieldType)
Schema    = List (TableName × TableType)

infixr 30 _::_
data Row : TableType -> Set where
  []   : Row []
  _::_ : forall {x a xs} -> Field a -> Row xs -> Row ((x , a) :: xs)

module AssocList {Key : Set}(_==_ : Key -> Key -> Bool) where

  hasKey : {A : Set} -> Key -> List (Key × A) -> Bool
  hasKey _ [] = false
  hasKey y ((x , _) :: xs) = x == y or hasKey y xs

  HasKey : {A : Set} -> Key -> List (Key × A) -> Set
  HasKey x xs = isTrue (hasKey x xs)

  lookup : {A : Set}(x : Key)(xs : List (Key × A)){p : HasKey x xs} -> A
  lookup _ [] {}
  lookup y ((x , v) :: xs) {p} with x == y
  ... | true  = v
  ... | false = lookup y xs {p}

  hasKeys : {A : Set} -> List Key -> List (Key × A) -> Bool
  hasKeys []        _ = true
  hasKeys (x :: xs) t = hasKey x t and hasKeys xs t

  HasKeys : {A : Set} -> List Key -> List (Key × A) -> Set
  HasKeys xs t = isTrue (hasKeys xs t)

  project : {A : Set}(xs : List Key)(tbl : List (Key × A)){p : HasKeys xs tbl} ->
            List (Key × A)
  project []        tbl     = []
  project (x :: xs) tbl {p} = (x , lookup x tbl {andElimL p}) ::
                              project xs tbl {andElimR (hasKey x tbl) p}

  distinctKeys : {A : Set} -> List (Key × A) -> Bool
  distinctKeys []              = true
  distinctKeys ((x , _) :: xs) = not (hasKey x xs) and distinctKeys xs 

  disjoint : {A : Set} -> List (Key × A) -> List (Key × A) -> Bool
  disjoint xs ys = distinctKeys (xs ++ ys)

  Disjoint : {A : Set} -> List (Key × A) -> List (Key × A) -> Set
  Disjoint xs ys = isTrue (disjoint xs ys)

open AssocList eqString

data ExprType : Set where
  bool  : ExprType
  ftype : FieldType -> ExprType

data Expr (t : TableType) : ExprType -> Set where
  _===_ : forall {a} -> Expr t a -> Expr t a -> Expr t bool
  !_    : forall x {p : HasKey x t} -> Expr t (ftype (lookup x t {p}))
  val   : forall {a} -> Field a -> Expr t (ftype a)

data Query (s : Schema) : TableType -> Set where
  read : (x : TableName){p : HasKey x s} -> Query s (lookup x s {p})
  _⊗_  : forall {t t'}{p : Disjoint t t'} -> Query s t -> Query s t' -> Query s (t ++ t')
  proj : forall {t} xs {p : HasKeys xs t} -> Query s t -> Query s (project xs t {p})
  select : forall {t} -> Expr t bool -> Query s t -> Query s t

Database = String
UserName = String
Password = String

postulate
  RawHandle : Set

data Handle (s : Schema) : Set where
  handle : RawHandle -> Handle s

postulate
  rawConnect   : Database -> UserName -> Password -> IO RawHandle
  rawGetSchema : RawHandle -> IO Schema
  sameSchema   : Schema -> Schema -> Bool

connect : Database -> UserName -> Password -> (s : Schema) -> IO (Handle s)
connect db user pass s =
  rawConnect db user pass >>= \h ->
  rawGetSchema h >>= \s' ->
  if sameSchema s s'
    then return (handle h)
    else ioError "schema mismatch"

postulate
  RawField : Set
  rawQuery : RawHandle -> String -> IO (List (List RawField))
  toSQL    : forall {s t} -> Query s t -> String

query : forall {s t} -> Handle s -> Query s t -> IO (List (Row t))
query (handle h) q = {! !}

studentTable : TableType
studentTable =
  ("Id" , number ) ::
  ("FirstName" , text 20) ::
  ("LastName"  , text 20) :: []

courseTable : TableType
courseTable =
  ("Code" , text 6)  ::
  ("Name" , text 20) :: []

registrationTable : TableType
registrationTable =
  ("CourseCode" , text 3) ::
  ("StudentId"  , number) :: []

schema : Schema
schema =
  ("Students" , studentTable) ::
  ("Courses"  , courseTable ) ::
  ("Registration" , registrationTable) :: []

findStudents : VString 3 -> Query schema _
findStudents code =
  proj ("FirstName" :: "LastName" :: [])
  ( select (! "CourseCode" === val code)
    ( select (! "StudentId" === ! "Id")
      (read "Students" ⊗ read "Registration")
    )
  )
\end{code}
